import json
import os
from functools import partial
from pathlib import Path
from voxl_elrs_tools.wizard import PyWizard, values, home, enum_select, clear
from voxl_elrs_tools.elrs_device import ELRSDeviceManager
from voxl_elrs_tools.crsf import CRSFDev, CRSFSetID, CRSFSetPWMCh, CRSFSetBootloader
from voxl_elrs_tools.manage_firmware import ELRSFirmwareManager
import time

class ELRSConfigurator:
    
    def __init__(self, dev: ELRSDeviceManager = None, conf_file = None) -> None:
        if dev is None:
            self.dev = ELRSDeviceManager()
        else:
            self.dev = dev

        if conf_file is None:
            self.conf_file = self.dev.conf_file
        else:
            self.conf_file = conf_file
        
        if not os.path.exists(self.conf_file):
            print("Configuration file does not exist, creating...")
            Path(self.conf_file).touch()
        elif not os.path.isfile(self.conf_file):
            raise FileExistsError("Configuration file is a directory")

        self.configuration = None

    def load_data(self, file_name=None):
        try:
            fp = open(self.conf_file if file_name is None else file_name, "r")
            conf_data = json.load(fp)
            self.configuration = conf_data
        except:
            self.configuration = None
        finally:
            fp.close()

        if file_name is not None:
            # Also copy it to default conf file
            self.save_data()

    def save_data(self, wizard=None, file_name=None):
        if file_name is not None:
            try:
                with open(file_name, "w") as fp:
                    json.dump(self.configuration, fp)
            except:
                pass
        else:
            try:
                with open(self.conf_file, "w") as fp:
                    json.dump(self.configuration, fp)
            except:
                pass
        # if wizard is not None:
        #     time.sleep(1) # Just to make sure the popup actually appears first since the save happens fast
        #     wizard.controller.stop_loading_popup()

    def get_configuration(self) -> dict:
        return self.configuration

    def set_configuration(self, conf: dict) -> None:
        self.configuration = conf

    def upload_configuration(self, conf_data: dict = None, try_update: bool = True, wizard = None) -> int:
        if conf_data is not None:
            conf = conf_data
        else:
            conf = self.configuration
        
        # First check model name (if necessary)
        dev_info = CRSFDev(self.dev).run()
        if conf["model_name"] is not None and len(conf["model_name"]) > 0 and dev_info is not None:
            if conf["model_name"] != dev_info["name"]:
                raise Exception("Configuration name mismatch")
                print("Model name does not match expected")
                if wizard is not None:
                    time.sleep(1)
                    wizard.controller.stop_loading_popup()
                return 1

        # First upload (if enabled)
        if try_update and dev_info is not None and (conf["version"] == "latest" or conf["version"] == ""):
            firm = ELRSFirmwareManager(self.dev)
            ret = firm.auto_update(name=dev_info["name"], min_ver=dev_info["sw_ver"])
            #dev_info = CRSFDev(self.dev).run()
        elif try_update and dev_info is not None and conf["version"] == "any":
            print("Skipping update...")
        elif try_update and dev_info is not None and conf["version"] != "latest":
            firm = ELRSFirmwareManager(self.dev)
            fw_struct = firm.get_firmware(name=dev_info["name"], version=conf["version"])
            ret = firm.upload_firmware_from_dict(fw_struct)
            #dev_info = CRSFDev(self.dev).run()
        else:
            print("Skipping update...")

        # Then set pwm channels
        print("Setting pwm channels...")
        for pin, pwm_data in enumerate(conf.get("pwm", [])):
            time.sleep(1)
            if pwm_data is not None and pwm_data["enabled"]:
                CRSFSetPWMCh(
                    pin=pin,
                    in_ch=pwm_data.get("ch", 254),
                    mode=pwm_data["mode"],
                    failsafe_mode=pwm_data["failsafe_mode"],
                    failsafe_val=pwm_data["failsafe"],
                    map_mode=pwm_data["curve_mode"],
                    map_vals=pwm_data["curve_points"],
                    dev_manager=self.dev).run()
            else:
                # CH 255 and mode ON/OFF
                CRSFSetPWMCh(pin=pin, in_ch=254, mode=7, map_mode=0, map_vals=pwm_data["curve_points"], dev_manager=self.dev).run()

        # Then set binding phrase
        if conf["binding_phrase"] is not None and len(conf["binding_phrase"]) > 0:
            print("Setting the binding phrase...")
            CRSFSetID(dev_manager=self.dev, binding_phrase=conf["binding_phrase"]).run()

        #print("Resetting...")
        #CRSFSetBootloader(dev_manager=self.dev).run()
        if wizard is not None:
            wizard.controller.stop_loading_popup()

        return 0
    
    def configure_default(self):
        self.load_data()
        if self.configuration is not None:
            self.upload_configuration()
        else:
            print("No configuration to upload")


def percent_to_pwm(percent: int) -> int:
    return int((percent/100) * 1024)

def pwm_to_percent(pwm_val: int) -> int:
    return int((pwm_val / 1024) * 100)

class ELRSConfigData:
    """
    Helper methods for serializing/deserializing configuration in JSON form and wizard form
    """

    # A default configuration dict
    default_data = {
        "Configuration name": "Custom",
        "Set PWM channels": {
            "PWM 0": {
                "Enabled": False,
                "Channel": 255,
                "Mode": 0,
                "Failsafe percentage": 50,
            },
            "PWM 1": {
                "Enabled": False,
                "Channel": 255,
                "Mode": 0,
                "Failsafe percentage": 50,
            },
            "PWM 2": {
                "Enabled": False,
                "Channel": 255,
                "Mode": 0,
                "Failsafe percentage": 50,
            },
            "PWM 3": {
                "Enabled": False,
                "Channel": 255,
                "Mode": 0,
                "Failsafe percentage": 50,
            },
        },
        "Require a receiver model": "",
        "Specify a version": "latest",
        "Choose a binding phrase": "Default ELRS binding phrase",
    }

    def __init__(self, json_data: str = None, conf_data: dict = None) -> None:
        # Data model
        self.conf_name = "Custom"
        self.pwm = [None, None, None, None]
        self.model_name = None
        self.version_str = "latest"
        self.binding_str = "Default ELRS binding phrase"

        # Load any initial data
        if json_data is not None:
            self.load_json(json_data)
        elif conf_data is not None:
            self.load_dict(conf_data)

    def load_json(self, json_str: str) -> None:
        data = json.loads(json_str)
        self.conf_name = data["name"]
        self.pwm = data["pwm"]
        self.binding_str = data["binding_phrase"]
        self.version_str = data["version"]
        self.binding_str = data["binding_phrase"]

    def load_json_dict(self, data: dict) -> None:
        self.conf_name = data["name"]
        self.pwm = data["pwm"]
        self.binding_str = data["binding_phrase"]
        self.version_str = data["version"]
        self.binding_str = data["binding_phrase"]

    def get_json_ready(self) -> dict:
        return {
            "name": self.conf_name,
            "pwm": self.pwm,
            "binding_phrase": self.binding_str,
            "version": self.version_str,
            "model_name": self.model_name
        }

    def print_json(self) -> str:
        return json.dumps(self.get_json_ready(), indent=1)

    def load_dict(self, conf: dict):
        self.conf_name = conf["Configuration name"]
        self.model_name = None if \
            conf["Require a receiver model"] is None or len(conf["Require a receiver model"]) < 1 else \
            conf["Require a receiver model"]
        self.version_str = conf["Specify a version"]
        self.binding_str = conf["Choose a binding phrase"]

        for k,v in conf["Set PWM channels"].items():
            ch_num = int(k[-1])
            self.pwm[ch_num] = None if not v["Enabled"] else {
                "ch": v["Channel"],
                "mode": v["Mode"],
                "failsafe": percent_to_pwm(v["Failsafe percentage"])
            }


    def get_dict(self) -> dict:
        new_dict = dict(ELRSConfigData.default_data)
        new_dict["Set PWM channels"] = {
            "PWM {}".format(index) : {
                "Enabled": False if data is None else True,
                "Channel": 255 if data is None else data["ch"],
                "Mode": 0 if data is None else data["mode"],
                "Failsafe percentage": 50 if data is None else pwm_to_percent(data["failsafe"])
            } for index, data in enumerate(self.pwm)
        }
        new_dict["Configuration name"] = self.conf_name
        new_dict["Require a receiver model"] = self.model_name
        new_dict["Specify a version"] = self.version_str
        new_dict["Choose a binding phrase"] = self.binding_str
        return new_dict
        
    def get_default_dict() -> dict:
        """
        Static method
        """
        return ELRSConfigData(conf_data=ELRSConfigData.default_data).get_dict()
    

##################################################################
### Configuration wizard commands/tools
##################################################################


def new_conf():
    conf = ELRSConfigData.get_default_dict()
    # Get config data from the wizard
    
    return partial(PyWizard.run_values_wizard, conf_struct, conf, "Set all values below:\n", False)


def edit_conf(name: str):
    obj = ELRSConfigData()
    obj.load_json_dict(ELRSConfigurator().get_configuration_by_name(name=name))
    conf = obj.get_dict()

    PyWizard.run_values_wizard(conf_struct, conf, "Set all values below:\n", False)


def choose_conf_name(post_exec: callable = None):
    conf_names = list(ELRSConfigurator().get_configurations().keys())

    name_index = PyWizard.render_opts(conf_names, "Choose a configuration:\n", "> ")
    if name_index != 0 and name_index <= len(conf_names):
        name = conf_names[name_index - 1]
    elif name_index == 0:
        return PyWizard.pop()

    if post_exec is not None:
        PyWizard.stack.append((choose_conf_name, (post_exec, ), ))
        post_exec(name)


def edit_configs():
    return choose_conf_name(edit_conf)


def delete_configs():
    return choose_conf_name(lambda name: ELRSConfigurator().delete_configuration(name))


def set_def_configs():
    print(f"Current default: {ELRSConfigurator().get_default()}")
    return choose_conf_name(lambda name: ELRSConfigurator().set_default(name))


def upload_from_list():
    return choose_conf_name(lambda name: ELRSConfigurator().upload_configuration(conf_name=name))


def pwm_values(arg: dict):
    renderer = lambda o, d: "{} : {}".format(o, "Enabled, Ch {}, Mode {}, Failsafe {}".format(d["Channel"], pwm_mode_enum[d["Mode"]], d["Failsafe percentage"]) if d["Enabled"] else "Disabled")
    getter = lambda data, key: PyWizard.run_values_wizard(arg, data[key], "Set pwm data\n", False)
    post = PyWizard.pop
    return renderer, getter, post


def write_conf(data=None, option=None):
    # Write the config to a file
    print(ELRSConfigData(conf_data=data).print_json())

    conf = ELRSConfigData(conf_data=data).get_json_ready()
    conf_manager = ELRSConfigurator()
    conf_manager.add_configuration(conf)

    # Upload the config (if the user wants)
    print("\nUpload this configuration to the receiver? (Y/N)")
    user = ""
    while user.lower() != "y" and user.lower() != "n":
        user = input("> ")
    if user.lower() == "y":
        conf_manager.upload_configuration(conf["name"])


##################################################################
### Configuration wizard definition
##################################################################


pwm_mode_enum = { 0: "Default", }
pwm_value_struct = {
    "Enabled": values(bool, ),
    "Channel": values(int, ),
    "Mode": enum_select(pwm_mode_enum),
    "Failsafe percentage": values(int, )
}

conf_struct = {
    "Configuration name": values(str, ),
    "Set PWM channels": {
        "PWM 0": pwm_values(pwm_value_struct),
        "PWM 1": pwm_values(pwm_value_struct),
        "PWM 2": pwm_values(pwm_value_struct),
        "PWM 3": pwm_values(pwm_value_struct),
    },
    "Require a receiver model": values(str, ),
#       TODO: Add nightly support
#       "Specify a version": values(str, "Choose a version string to match, use \'latest\' to always use the most recent release, or \'nightly\' to use the latest beta image"),
    "Specify a version": values(str, "Choose a version string to match or enter \'latest\' to always use the most recent release"),
    "Choose a binding phrase": values(str, ),
    "Confirm": write_conf
}

## Config wizard prompt and structure
config_prompt = """Please choose a configuration option

==============================
"""
config_wizard = {
    "Upload configuration to receiver": home(upload_from_list),
    "Create new configuration": home(new_conf()),
    "Edit configuration": home(edit_configs),
    "Delete configuration": home(delete_configs),
    "Set configuration as default": home(set_def_configs),
}

if __name__ == "__main__":
    c = ELRSConfigurator()
    c.load_data()
    print(c.get_configuration())