# Adapted from https://github.com/crsf-wg/crsf/wiki/Python-Parser
import math
import sys
import time

from voxl_elrs_tools import crc
import ctypes
from ctypes import c_uint8 as uint8_t
from typing import Union, Tuple
from enum import IntEnum, Enum
from voxl_elrs_tools.elrs_device import ELRSDeviceManager
import hashlib

SYNC_BYTE = 0xC8

CRSF_CHANNEL_VALUE_MIN = 172
CRSF_CHANNEL_VALUE_1000 = 191
CRSF_CHANNEL_VALUE_MID = 992
CRSF_CHANNEL_VALUE_2000 = 1792
CRSF_CHANNEL_VALUE_MAX = 1811

### Helpers for converting values to CRSF channel values

def map_value(value, original_min, original_max, target_min, target_max):
    return target_min + (value - original_min) * (target_max - target_min) / (original_max - original_min)

def std_to_crsf(in_val):
    """
    Converts values in the range [-100, 100] to values in the range [172, 1811]
    """
    return int(map_value(in_val, -100, 100, CRSF_CHANNEL_VALUE_1000, CRSF_CHANNEL_VALUE_2000))

def us_to_crsf(in_val):
    return int(map_value(in_val, 1000, 2000, CRSF_CHANNEL_VALUE_1000, CRSF_CHANNEL_VALUE_2000))

def crsf_to_std(in_val):
    return int(map_value(in_val, CRSF_CHANNEL_VALUE_1000, CRSF_CHANNEL_VALUE_2000, -100, 100))

def crsf_to_us(in_val):
    return int(map_value(in_val, CRSF_CHANNEL_VALUE_1000, CRSF_CHANNEL_VALUE_2000, 1000, 2000))


class PacketsTypes(IntEnum):
    """
    CRSF Packet Types
    """
    GPS = 0x02
    VARIO = 0x07
    BATTERY_SENSOR = 0x08
    BARO_ALT = 0x09
    HEARTBEAT = 0x0B
    VIDEO_TRANSMITTER = 0x0F
    LINK_STATISTICS = 0x14
    RC_CHANNELS_PACKED = 0x16
    ATTITUDE = 0x1E
    FLIGHT_MODE = 0x21
    DEVICE_INFO = 0x29
    CONFIG_READ = 0x2C
    CONFIG_WRITE = 0x2D
    RADIO_ID = 0x3A
    COMMAND = 0x32
    DEVICE_PING = 0x28
    NO_PACKET = 0x0

class SubCommandTypes(Enum):
    """
    CRSF Packet Types
    """
    BIND = bytes((0x62, 0x64))
    UNBIND = bytes((0x75, 0x62))
    ID = bytes((0x69, 0x64))
    PWM_CH = bytes((0xF3, ))
    PWM_VAL = bytes((0xF4, ))
    BOOTLOADER = bytes((0x62, 0x6C))


class RCChannelsPacked(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("channel_01", ctypes.c_int, 11),
        ("channel_02", ctypes.c_int, 11),
        ("channel_03", ctypes.c_int, 11),
        ("channel_04", ctypes.c_int, 11),
        ("channel_05", ctypes.c_int, 11),
        ("channel_06", ctypes.c_int, 11),
        ("channel_07", ctypes.c_int, 11),
        ("channel_08", ctypes.c_int, 11),
        ("channel_09", ctypes.c_int, 11),
        ("channel_10", ctypes.c_int, 11),
        ("channel_11", ctypes.c_int, 11),
        ("channel_12", ctypes.c_int, 11),
        ("channel_13", ctypes.c_int, 11),
        ("channel_14", ctypes.c_int, 11),
        ("channel_15", ctypes.c_int, 11),
        ("channel_16", ctypes.c_int, 11),
    ]

class LinkStatistics(ctypes.Structure):
    _pack_ = 1  # match the packed attribute in C
    _fields_ = [
        ("up_rssi_ant1",    ctypes.c_uint8),
        ("up_rssi_ant2",    ctypes.c_uint8),
        ("up_link_quality", ctypes.c_uint8),
        ("up_snr",          ctypes.c_int8),
        ("active_antenna",  ctypes.c_uint8),
        ("rf_profile",      ctypes.c_uint8),
        ("up_rf_power",     ctypes.c_uint8),
        ("down_rssi",       ctypes.c_uint8),
        ("down_link_quality", ctypes.c_uint8),
        ("down_snr",        ctypes.c_int8),
    ]

packet_type_mapping = {
    PacketsTypes.RC_CHANNELS_PACKED: RCChannelsPacked,
    PacketsTypes.LINK_STATISTICS: LinkStatistics,
}

class RxPwmConfigIn(ctypes.LittleEndianStructure):
    _pack_ = 1  # Make sure fields are laid out without extra padding
    _fields_ = [
        ("pinIndex",      ctypes.c_uint8),
        ("failsafe",      ctypes.c_uint16),
        ("inputChannel",  ctypes.c_uint8),
        ("inverted",      ctypes.c_uint8),
        ("mode",          ctypes.c_uint8),
        ("narrow",        ctypes.c_uint8),
        ("failsafeMode",  ctypes.c_uint8),
        ("mapMode",       ctypes.c_uint8),
        ("unused",        ctypes.c_uint8),
        ("mapInVal1",     ctypes.c_uint16),
        ("mapInVal2",     ctypes.c_uint16),
        ("mapInVal3",     ctypes.c_uint16),
        ("mapOutVal1",    ctypes.c_uint16),
        ("mapOutVal2",    ctypes.c_uint16),
        ("mapOutVal3",    ctypes.c_uint16),
    ]

def str2us(microseconds: int) -> Tuple[uint8_t, uint8_t]:
    if microseconds < 256:
        return microseconds, 0
    else:
        return (microseconds >> 8),(microseconds & 0xFF)

def generateUID(phrase):
    uid = [
        int(item) if item.isdigit() else -1
        for item in phrase.split(',')
    ]
    if (4 <= len(uid) <= 6) and all(ele >= 0 and ele < 256 for ele in uid):
        # Extend the UID to 6 bytes, as only 4 are needed to bind
        uid = [0] * (6 - len(uid)) + uid
        uid = bytes(uid)
    else:
        uid = hashlib.md5(("-DMY_BINDING_PHRASE=\""+phrase+"\"").encode()).digest()[0:6]
    return uid

class CRSFCommand:
    """
    Standard interface for sending a crossfire command and reading the response.
    Most routines can be ran easiest by creating the object and calling run() on it.

    Methods
    ----------
    - send: Send the prepared command
    - recv: Receive any response to the command
    - run: (Abstract) Run the command and process any response
    - parse_response

    Attributes
    ----------
    """

    def __init__(self, dev_manager: ELRSDeviceManager = None, cmd: bytearray = None, cmd_type: PacketsTypes = None, subcmd_type: SubCommandTypes = None, payload: bytearray = None) -> None:
        self.cmd: bytearray = cmd
        if cmd is not None:
            self.has_crc = True
        else:
            self.has_crc = False
        if dev_manager is None:
            self.dev_manager = ELRSDeviceManager()
            self.device = self.dev_manager.open_device()
        else:
            self.dev_manager = dev_manager
            if self.dev_manager.dev is not None:
                self.device = self.dev_manager.dev
            else:
                self.device = self.dev_manager.open_device()

        if self.cmd is None:
            if cmd_type is not None:
                self.gen_command_header(cmd_type, subcmd_type)
                if payload is not None:
                    self.cmd.extend(payload)
                self.set_meta()

    def send(self) -> int:
        if self.device is not None and self.cmd is not None:
            if self.dev_manager.dev is None:
                self.dev_manager.open_device()
            try:
                self.dev_manager.dev.write(bytearray(self.cmd))
            except:
                return -1
        else:
            raise Exception("Device and/or command not set!")
        
        return 0

    def recv(self) -> Tuple[PacketsTypes, bytes]:
        data = self.device.read(64)
        try:
            ptype, response = self.parse_response(data)
        except:
            return PacketsTypes.NO_PACKET, None
        if ptype != -1:
            return ptype, response
        else:
            return PacketsTypes.NO_PACKET, None

    def parse_response(self, input, expected_pkt: PacketsTypes=None) -> Tuple[PacketsTypes, Union[bytearray, None]]:
        while len(input) > 4:
            val = input[0]
            if val == SYNC_BYTE:
                size = input[1]
                payload = input[2:3 + size]
                if payload[-1] == SYNC_BYTE: payload = payload[:-1]
                packet_crc = payload[-1]
                if packet_crc == crc.calc(payload[:-1]):
                    packet_type: PacketsTypes = PacketsTypes(input[2])
                    if expected_pkt is None or expected_pkt == packet_type:
                        return (packet_type, payload[1:-1])
            input = input[1:]
        return (-1, None)
    
    def gen_command_header(self, cmd_type : PacketsTypes, subcmd_type: SubCommandTypes = None):
        data_out = bytearray()
        data_out.append(0xEC)  # ADDR -> RECEIVER
        data_out.append(0x00)
        data_out.append(cmd_type.value)
        if cmd_type == PacketsTypes.COMMAND:
            if subcmd_type is not None:
                data_out.extend(subcmd_type.value)
        self.cmd = data_out


    def add_crc(self) -> None:
        self.cmd.append(crc.calc(self.cmd[2:]))
        self.has_crc = True

    def set_len(self) -> None:
        self.cmd[1] = len(self.cmd) - 2 if self.has_crc else len(self.cmd) - 1

    def set_meta(self) -> None:
        self.set_len()
        self.add_crc()

    def run(self):
        raise NotImplementedError()


class CRSFBind(CRSFCommand):
    """
    Put the receiver into binding mode
    """

    def __init__(self, dev_manager=None) -> None:
        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.BIND)
    
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()


class CRSFDev(CRSFCommand):

    def __init__(self, dev_manager=None) -> None:
        payload = bytes((0xEC, 0xC8, 0xEC))
        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.DEVICE_PING,
            payload=payload)

    def recv(self) -> Tuple[PacketsTypes, bytes]:
        ptype = -1
        count = 1
        response = None
        while ptype == -1 and count > 0:
            data = self.device.read(64)
            count -= 1
            ptype, response = self.parse_response(data, expected_pkt=PacketsTypes.DEVICE_INFO)
        if ptype != -1:
            return response
        else:
            return None

    def parse_info(self, response) -> dict:
        _device_name:str =""
        _sw_version:str =""

        for b in response[2:]:
            if b == 0:
                break
            _device_name += chr(b)
    
        # Get Version -> Version is always the [-7:-3] bytes in the packet we receive in the current version of ELRS
        _sw_version = '.'.join([str(int(b)) for b in response[-6:-2]])
        if _sw_version[0] == "0":
            _sw_version = _sw_version[2:]
        _hw_version = str(response[-10:-6]) if int.from_bytes(response[-10:-6], "big") != 0 else "ELRS"
        _serial_num = str(response[-14:-10], "utf-8")
        if _serial_num != 'ELRS':
            print(f"Failed to validate version info: {_serial_num} != \"ELRS\"")

        return {
            "name": _device_name,
            "sw_ver": _sw_version,
            "hw_ver": _hw_version,
            "serial_num": _serial_num,
            "modal_elrs": not _hw_version == "ELRS", # True if _hw_version is set to m184 or other device
        }

    def format_info(self, info):
        return f"{info['name']}:{info['sw_ver']}"
        

    def run(self, retries=3) -> dict:
        for _ in range(retries):
            self.send()
            response = self.recv()
            if response is not None:
                break
        if response is not None:
            return self.parse_info(response)
        else:
            return None


class CRSFUnbind(CRSFCommand):
    """
    Unbind the receiver
    """

    def __init__(self, dev_manager=None) -> None:
        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.UNBIND)
    
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()


class CRSFSetID(CRSFCommand):
    """
    Sets the reciever's UID based on a binding phrase or directly
    """

    def __init__(self, binding_phrase : str = None, new_id : bytes = None, dev_manager=None, cmd=None) -> None:
        if new_id is not None:
            payload = new_id
        elif binding_phrase is not None:
            payload = generateUID(binding_phrase)
        else:
            raise ValueError("No binding phrase or id")

        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.ID,
            payload=payload)
    
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()


class CRSFSetPWMCh(CRSFCommand):
    """
    Set the input channel a pwm pin should listen on
    """

    def __init__(
            self,
            pin: int,
            in_ch: int,
            inverted=False,
            mode=0,
            narrow=0,
            failsafe_mode=0,
            failsafe_val=0,
            map_mode=0,
            map_vals=[],
            dev_manager=None) -> None:
        chan_struct = RxPwmConfigIn(
            pinIndex = pin,
            inputChannel = in_ch - 1,
            inverted = 1 if inverted else 0,
            mode = mode,
            narrow = narrow,
            failsafeMode = failsafe_mode,
            failsafe = max(failsafe_val - 988, 0),
            unused=0,
            mapMode = map_mode,
            mapInVal1 = std_to_crsf(map_vals[0][0]),
            mapInVal2 = std_to_crsf(map_vals[1][0]),
            mapInVal3 = std_to_crsf(map_vals[2][0]),
            mapOutVal1 = map_vals[0][1],
            mapOutVal2 = map_vals[1][1],
            mapOutVal3 = map_vals[2][1],
        )
        payload = bytes(chan_struct)
        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.PWM_CH,
            payload=payload)
    
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()

class CRSFOverridePWM(CRSFCommand):
    """

    """

    def __init__(self, rc_ch: int, ch_val: int, dev_manager=None):
        payload = bytes((0xEC, 0xC8, 0xF4, rc_ch, (ch_val & 0xFF00) >> 8, ch_val & 0xFF))
        super().__init__(dev_manager, cmd_type=PacketsTypes.COMMAND, subcmd_type=None, payload=payload)

    def recv(self) -> bytes:
        return None

    def run(self) -> int:
        return self.send()


class CRSFSetBootloader(CRSFCommand):
    """
    Jump back to the bootloader for restarting the device or uploading firmware
    """

    def __init__(self, dev_manager=None) -> None:
        super().__init__(
            dev_manager,
            cmd_type=PacketsTypes.COMMAND,
            subcmd_type=SubCommandTypes.BOOTLOADER)
    
    def recv(self) -> bytes:
        return None
    
    def run(self) -> int:
        return self.send()

class CRSFReader(CRSFCommand):
    def __init__(self, dev_manager=None) -> None:
        super().__init__(dev_manager)

    def get_packet(self, packet_type=None, timeout=5):
        start_time = time.time()
        ptype, response = self.recv()
        while (ptype == PacketsTypes.NO_PACKET or ((packet_type is not None) and (ptype != packet_type))) and time.time() - start_time < timeout:
            ptype, response = self.recv()
        return ptype, response

    def get_packet_data(self, packet_type, timeout=5):
        ptype, response = self.get_packet(packet_type, timeout=timeout)
        if ptype == PacketsTypes.NO_PACKET:
            return None
        return packet_type_mapping[ptype].from_buffer_copy(response)

    def parse_packet_data(self, ptype, response):
        if ptype == PacketsTypes.NO_PACKET:
            return None
        return packet_type_mapping[ptype].from_buffer_copy(response)

    def get_field(self, packet_type, field_attr, timeout=5):
        data = self.get_packet_data(packet_type, timeout=timeout)
        if data is None:
            return 0
        return data.__getattribute__(field_attr)

def send_tx_bind():
    dev = ELRSDeviceManager(baud = 400000)
    cmd_payload = bytearray((0xC8, 0x06, 0x32, 0xEE, 0xEA, 0x10, 0x01))
    cmd = CRSFCommand(dev_manager=dev, cmd=cmd_payload)
    cmd.add_crc()
    print(cmd.cmd.hex())
    cmd.send()

def plot_val(cmd: CRSFReader, packet_type: PacketsTypes, field: str):
    from matplotlib import pyplot as plt
    plt.ion()
    fig, ax = plt.subplots()
    xdata, ydata, data2, data3 = [], [], [], []
    (line,) = ax.plot(xdata, ydata, '-')
    (line2,) = ax.plot(xdata, ydata, '-')
    (line3,) = ax.plot(xdata, ydata, '-')

    while True:
        vals = cmd.get_packet_data(packet_type)
        if vals is None:
            continue
        val = vals.__getattribute__(field)

        t = time.time() - start_time
        xdata.append(t)
        ydata.append(val)
        data2.append(-vals.__getattribute__("up_rssi_ant1"))
        data3.append(-vals.__getattribute__("up_rssi_ant2"))

        # Update the line data
        line.set_xdata(xdata)
        line.set_ydata(ydata)
        # Update the line data
        line2.set_xdata(xdata)
        line2.set_ydata(data2)
        # Update the line data
        line3.set_xdata(xdata)
        line3.set_ydata(data3)

        # Rescale the axes to fit new data
        ax.relim()
        ax.autoscale_view()

        # Redraw the figure
        plt.draw()

        plt.pause(0.01)

    # Keep the plot open (or comment this out if you want it to close at EOF)
    plt.ioff()
    plt.show()

if __name__ == "__main__":
    #dev = CRSFSetID(binding_phrase="modal dbg phrase")
    # start_time = time.time()
    # dev = ELRSDeviceManager(baud = 420000)
    # cmd = CRSFReader(dev_manager=dev)
    #
    # plot_val(cmd, PacketsTypes.LINK_STATISTICS, "up_link_quality")

    # print(cmd.get_field(PacketsTypes.LINK_STATISTICS, "up_rssi_ant2"))

    dev = ELRSDeviceManager(baud=420000)
    c = CRSFOverridePWM(7, 191, dev_manager=dev)
    print(c.cmd.hex().upper())
    c.send()