import ctypes as ct
import sys
import os
import os.path as pt
import re
import subprocess
import serial
import time
import threading

from voxl_utils_python.voxl_utils import is_voxl
from .px4_utils import toggle_voxl_px4, px4_running, drone_armed

elrs_dir = '/usr/share/modalai/voxl-elrs/'

elrs_default_device = None
if is_voxl():
    elrs_default_device = '/dev/tty-slpi-7'
else:
    elrs_default_device = '/dev/ttyUSB0'


def device_manager_thread(device):
    """
    This method is used to manage a uart connection in the background, making sure that the device stays open and
    keeps its ringbuffer from getting full

    :param device: `ELRSDeviceManager` - The device to manage
    """
    while not device.stop_uart_event.is_set():
        time.sleep(1)
        device.uart_lock.acquire()
        if is_voxl() and device.dev is not None:
            device.dev.flush()
        device.uart_lock.release()

class ELRSDeviceManager:
    """
    Object containing the UART interface between the ELRS device and the host
    """

    def __init__(self, device=None, baud=420000) -> None:
        self.is_voxl = is_voxl()
        self.device = None
        self.dev = None
        self.baud = baud

        if device is not None:
            self.device = device

        # Detect if we are on a voxl2 or host computer
        if self.is_voxl:
            base_dir = elrs_dir
            self.host = "voxl2"
            self.enabled = None
            self.running = None
        else:
            base_dir = elrs_dir
            self.host = "ubuntu"

        # Directory paths
        self.RX_ESPTOOL = self.TX_ESPTOOL = pt.join(base_dir, "esptool/esptool.py")
        self.TOOLS_DIR = pt.join(base_dir, "tools")
        # Legacy directories for binaries distributed with the package
        # All commands using these should eventually be replaced
        self.TX_DIR = pt.join(base_dir, "firmware/binaries/tx")
        self.RX_DIR = pt.join(base_dir, "firmware/binaries/rx")
        self.FIRMWARE_DIR = pt.join(base_dir, "firmware")
        self.TX_3_0_1 = [self.TX_DIR + "/bootloader_dio_40m.bin", self.TX_DIR + "/partitions.bin", self.TX_DIR + "/boot_app0.bin",  self.TX_DIR + "/iFlight_900_TX-3.0.1.bin"]
        self.TX_V2_3_4_2 = [self.TX_DIR + "/bootloader_dio_40m_v2.bin", self.TX_DIR + "/partitions_v2.bin", self.TX_DIR + "/boot_app0_v2.bin",  self.TX_DIR + "/iFlight_900_TX_V2-3.4.2.bin"]
        self.TX_M0193 = self.FIRMWARE_DIR + "/3.5.3.4/MODALAI_M0193_TX_via_UART/ModalAI_M0193_TX-3.5.3.4.bin"
        self.conf_file = pt.abspath("/etc/modalai/voxl-elrs.conf")
        self.base_dir = base_dir
        sys.path.append(self.TOOLS_DIR)

        self.thread = None
        self.uart_lock = threading.Lock()
        self.stop_uart_event = threading.Event()

    def start_management_thread(self):
        self.thread = threading.Thread(target=device_manager_thread, args=(self, ), daemon=True)
        self.thread.start()

    def open_device(self, device=None):
        if device is None:
            if self.device is not None:
                device = self.device
            else:
                device = elrs_default_device

        if self.is_voxl:
            en, run = toggle_voxl_px4()
            if self.enabled is None:
                self.enabled, self.running = en, run
            #subprocess.run("voxl-reset-slpi")

        if self.is_voxl:
            from voxl_utils_python.voxl_serial import VoxlSerialPort
            dev = VoxlSerialPort()
            dev.port = device
            dev.baudrate = self.baud
            dev.open()
        else:
            dev = serial.Serial(device, self.baud, timeout=1, parity=serial.PARITY_NONE, stopbits=serial.STOPBITS_ONE, bytesize=serial.EIGHTBITS)

        self.device = device
        self.dev = dev
        return dev

    def close(self, restart_px4=True, close_fds=False):
        if self.dev is not None:
            self.dev.close()
            self.dev = None
            if close_fds and is_voxl():
                try:
                    os.closerange(3,10)
                    pass
                except:
                    pass
        if self.is_voxl and self.enabled is not None and restart_px4:
            toggle_voxl_px4(enabled=self.enabled, running=self.running)

    def __del__(self):
        self.close()