#!/usr/bin/python3
import subprocess
import time
import os
import re
import subprocess

# COLORS!
RESET = "\033[0m"
GRN = "\033[1;32m"
LIT_GRN = "\033[92m"
RED = "\033[1;31m"
WARN = "\033[33m"
CLEAR ="\033c"
BOLD="\033[1m"
RESET_BOLD="\033[21m"

VOXL_PLATFORM = False
try:
	ret = subprocess.check_output(['voxl-platform'])
	VOXL_PLATFORM = True
except FileNotFoundError:
	VOXL_PLATFORM = False

# Detect if we are on a voxl2 or host computer
if VOXL_PLATFORM:
	BASE_DIR = '/usr/share/modalai/voxl-elrs'
	host = "voxl2"
else:
	BASE_DIR = '/usr/share/modalai/voxl-elrs'
	host = "ubuntu"

# print(BASE_DIR)

RX_DIR = BASE_DIR + "/esptool/rx"

# BETAFPV_RX = "/usr/share/modalai/voxl-elrs/esptool/rx/BETAFPV_900_RX-3.2.1.2.bin"
# Firmware .bin paths for RX and TX
for root, dirs, files in os.walk(RX_DIR):
	for file in files:
		if re.search("betafpv", file, re.IGNORECASE):
			BETAFPV_RX = os.path.abspath(os.path.join(root, file))

# Check if a service is running
def service_running(service_name):
    """
    Checks if a systemd service is running or not

    @service_name: name of service to check
    @return: True/False depending on service state
    """
    cmd = ['systemctl', 'is-active', service_name]
    result = subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    if result.returncode == 0:
        return True
    else:
        return False
    
# Toggle voxl-px4 on or off
def toggle_voxl_px4(enabled=None, running=None):
	"""
	Toggle the state of voxl-px4. Needs to be off when interactive with the receiver.

	@running: Running state of voxl-px4
	@return: Returns the state of voxl-px4.
	"""
	
	start_voxl_px4_command = ['systemctl', 'start', 'voxl-px4']
	stop_voxl_px4_command = ['systemctl', 'stop', 'voxl-px4']

	# Case when we want to just stop voxl-px4
	if running == None:
		if service_running("voxl-px4"):
			# _print(f"[INFO] Stopping voxl-px4.", QUIET)
			running = True
			subprocess.run(stop_voxl_px4_command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
		else:
			running = False
	elif running:
		# _print(f"[INFO] Restarting voxl-px4.", QUIET)
		subprocess.run(start_voxl_px4_command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
	elif running == False:
		pass
		# _print(f"[INFO] Leaving voxl-px4 disabled.", QUIET)

	return enabled, running

# Verify user choice
def verify_choice():
	"""
	Verify the user wants to erase the application currently loaded onto a receiver.
	"""
	while True:
		print(f"{BOLD}{WARN}[WARNING] This operation is irreversible and will erase the application currently loaded. Do you want to continue? (y/n) {RESET}")
		ans = input().lower()
		if ans == "n":
			return 1
		elif ans == "y":
			return 0
		else:
			print(f"Invalid answer provided, please select a valid option (y/n).\n")
			continue

# Recover a receiver that is bricked
def recover_receiver():
	"""
	Recover an ELRS receiver that is bricked.
	
	@return: Fail or success
	"""

	options = ["Quit","BetaFPV Nano RX", "FrSky R9MM/Mini"]
	cases = {}
	# Query user for which receiver is being recovered
	while True:
		print(f"{LIT_GRN}Select an option:{RESET}")
		for index, option in enumerate(options):
			print(f"[{index + 1}] {option}")
			cases[index + 1] = option
		
		try:
			target = cases.get(int(input()))
			print(f"{CLEAR}", end="")
			break
		except:
			print(f"{WARN}Invalid answer provided, please select a valid option.{RESET}")
			continue
	
	if target == "BetaFPV Nano RX":
		# Verify user choice
		if verify_choice():
			return 0

		# Reboot to bootloader 
		print("[INFO] Power cycle your BetaFPV Nano RX and hold down the bootloader button while powering on the receiver. If done successfully, the LED on the receiver should be a solid blue light.")
		print(f"{GRN}Press enter to continue once the receiver is in bootloader mode.{RESET}")
		input()

		# Erase application 
		os.chdir("/usr/share/modalai/voxl-elrs/esptool")
		enabled, running = toggle_voxl_px4()
		time.sleep(0.1)
		print("[INFO] Erasing application...")
		result = subprocess.run("python3 esptool.py --port /dev/slpi-uart-7 --before no_reset erase_flash".split(), stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
		print(result.stdout.decode())
		if result.returncode > 1:
			print(f"{RED}[ERROR] Failed to erase chip. Double check that voxl-px4 is not running and that receiver is in bootloader mode.{RESET}")
			return -1
		
		# Flash application
		time.sleep(1)
		flash = input(f"{GRN}[INFO] Chip successfully erased, would you like to flash the latest firmware for BetaFPV Nano RX? (y/n){RESET}")
		if (flash.lower() == "y" ):
			print(f"[INFO] Flashing {target}...")
			print(f"[INFO] Flashing {BETAFPV_RX.split('/')[-1]}")
			result = subprocess.run(f"python3 esptool.py -b 460800 -p /dev/slpi-uart-7 -c esp8266 --before no_reset --after soft_reset write_flash 0x00 {BETAFPV_RX}".split(), stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
			print(result.stdout.decode())
			toggle_voxl_px4(enabled, running)
			if result.returncode:
				print(f"{RED}[ERROR] Failed to flash receiver. Double check that voxl-px4 is not running and that receiver is in bootloader mode.{RESET}")
				return -1
			else:
				print(f"{GRN}[INFO] Successfully recovered {target}!")
				return 0
		else:
			print(f"{GRN}[INFO] Skipping flashing receiver with latest firmware.{RESET}")
			print(f"{GRN}[INFO] Recovering receiver complete. No firmware loaded.{RESET}")

	elif target == "FrSky R9MM/Mini":
		# Verify user choice
		# if verify_choice():
			# return 0
		
		print("[INFO] Recovery for R9MM/Mini currently not supported through voxl-elrs.") 
		return -1
	
	elif target == "Quit":
		return 0
	
if __name__ == "__main__":
	exit(recover_receiver())
