import py_cui
import py_cui.widgets
from typing import Callable, Optional


class TwoDSliderWidget(py_cui.widgets.Widget):
    """
    A simple 2D 'joystick' or 'slider' widget that lets the user
    move a cursor in an X,Y plane with arrow keys, clamping
    values to [x_min..x_max] and [y_min..y_max].
    """

    def __init__(
            self, widget_id, title, grid, row, column, row_span, column_span,
            padx, pady, logger,
            x_min=0, x_max=100,
            y_min=0, y_max=100,
            init_x=50, init_y=50,
            x_step=1, y_step=1
    ):
        super().__init__(widget_id, title, grid, row, column, row_span, column_span, padx, pady, logger)
        # Range
        self._x_min, self._x_max = x_min, x_max
        self._y_min, self._y_max = y_min, y_max

        # Current values
        self._x_val = max(min(init_x, x_max), x_min)
        self._y_val = max(min(init_y, y_max), y_min)

        # Step sizes
        self._x_step = x_step
        self._y_step = y_step

        # Optional callback if you want to do something whenever value changes
        self._on_value_changed: Optional[Callable[[float, float], None]] = None

        # We want the user to be able to navigate with arrow keys:
        self.add_key_command(py_cui.keys.KEY_UP_ARROW, self._move_up)
        self.add_key_command(py_cui.keys.KEY_DOWN_ARROW, self._move_down)
        self.add_key_command(py_cui.keys.KEY_LEFT_ARROW, self._move_left)
        self.add_key_command(py_cui.keys.KEY_RIGHT_ARROW, self._move_right)

    def set_on_value_changed(self, callback: Callable[[float, float], None]):
        """
        Register a function to be called whenever x_val or y_val changes.
        The function signature should be:  def callback(x: float, y: float) -> None
        """
        self._on_value_changed = callback

    def _move_up(self):
        # Decrement or increment y, depending on your coordinate system
        if self._y_val < self._y_max:
            self._y_val += self._y_step
            self._notify_change()

    def _move_down(self):
        if self._y_val > self._y_min:
            self._y_val -= self._y_step
            self._notify_change()

    def _move_left(self):
        if self._x_val > self._x_min:
            self._x_val -= self._x_step
            self._notify_change()

    def _move_right(self):
        if self._x_val < self._x_max:
            self._x_val += self._x_step
            self._notify_change()

    def _notify_change(self):
        if self._on_value_changed is not None:
            self._on_value_changed(self._x_val, self._y_val)

    def get_x_val(self):
        return self._x_val

    def get_y_val(self):
        return self._y_val

    def set_x_val(self, x_val):
        self._x_val = max(min(x_val, self._x_max), self._x_min)
        self._notify_change()

    def set_y_val(self, y_val):
        self._y_val = max(min(y_val, self._y_max), self._y_min)
        self._notify_change()

    def _draw(self):
        """
        Renders a small ASCII box, e.g. 5x3, with a character for the cursor.
        The actual size in the terminal depends on how big you let py-cui
        draw this widget, so you might want to handle bigger or smaller dims.
        """

        super()._draw()
        if self._renderer is None:
            return

        self._renderer.set_color_mode(self._color)
        self._renderer.draw_border(self)

        # Calculate interior width/height (excluding borders/padding).
        height = self._height - 2 * self._pady - 2
        width = self._width - 2 * self._padx - 2 - 1
        base_y = self._start_y + 1

        if width <= 0 or height <= 0:
            self._renderer.unset_color_mode(self._color)
            return

        # Fill a 2D array with '.' as the background.
        matrix = [["." for _ in range(width)] for _ in range(height)]

        # Convert (x_val, y_val) to a row/col index for our ASCII grid.
        if self._x_max == self._x_min:
            cursor_col = 0
        else:
            ratio_x = (self._x_val - self._x_min) / (self._x_max - self._x_min)
            cursor_col = int(round(ratio_x * (width - 1)))

        if self._y_max == self._y_min:
            cursor_row = 0
        else:
            ratio_y = (self._y_val - self._y_min) / (self._y_max - self._y_min)
            # Invert so that higher y_val is drawn higher up in the box
            cursor_row = height - 1 - int(round(ratio_y * (height - 1)))

        # Build the string to display, e.g. "30,50" or with decimals "30.2,50.7"
        coord_str = f"{int(self._x_val)},{int(self._y_val)}"
        text_len = len(coord_str)

        # If we can fit coord_str starting at cursor_col, do so;
        # otherwise push as far right as possible while still fully visible.
        if cursor_col + text_len <= width:
            left_pos = cursor_col
        else:
            left_pos = width - text_len

        # Clamp left_pos to 0 in case the widget is too narrow.
        if left_pos < 0:
            left_pos = 0
            # Optionally truncate coord_str if it won't fit at all:
            # coord_str = coord_str[-width:]
            # text_len = len(coord_str)

        # Only draw if the cursor is in range.
        if 0 <= cursor_row < height:
            for i, ch in enumerate(coord_str):
                if 0 <= (left_pos + i) < width:
                    matrix[cursor_row][left_pos + i] = ch

        # Render each row of our ASCII matrix.
        for row_idx in range(height):
            line_str = "".join(matrix[row_idx])
            self._renderer.draw_text(self, line_str, base_y + row_idx, start_pos=0)

        # Optionally, show numeric x,y on the line below the box (if there's room).
        label_y = base_y + height
        if label_y < self._start_y + self._height - 1:
            label_str = f"X={self._x_val}, Y={self._y_val}"
            self._renderer.draw_text(self, label_str, label_y, start_pos=0)

        self._renderer.unset_color_mode(self._color)