import py_cui
import py_cui.widgets
import py_cui.ui

class SparklineCurveWidget(py_cui.widgets.Widget):
    """
    Displays an ASCII sparkline for up to three (x,y) points, within fixed bounds.
    Clamps out-of-range points. Shows basic axis labels on the bottom
    and prints a small label for each point's y-value.
    """

    def __init__(self, widget_id, title, grid, row, column, row_span, column_span,
                 padx, pady, logger,
                 points=None, mode="interpolate",
                 x_min=0, x_max=100, y_min=0, y_max=100):
        super().__init__(widget_id, title, grid, row, column, row_span, column_span, padx, pady, logger)
        self.set_selectable(False)  # Typically read-only
        self.mode = mode            # "step" or "interpolate"
        if points is None:
            points = [(10, 10), (50, 70), (80, 20)]
        self.points = points

        # We will fix bounding box for the graph
        self._x_min = x_min
        self._x_max = x_max
        self._y_min = y_min
        self._y_max = y_max

    def set_points(self, points):
        self.points = points

    def set_mode(self, mode):
        self.mode = mode

    def set_bounds(self, x_min, x_max, y_min, y_max):
        self._x_min = x_min
        self._x_max = x_max
        self._y_min = y_min
        self._y_max = y_max

    def _draw(self):
        super()._draw()
        if self._renderer is None:
            return

        self._renderer.set_color_mode(self._color)
        self._renderer.draw_border(self)

        # Dimensions inside the widget (excluding borders/padding)
        height = self._height - 2*self._pady - 2
        width  = self._width  - 2*self._padx  - 2
        base_y = self._start_y + 2  # skip top border
        base_x = self._start_x + 1  # skip left border

        # If no space to render, bail
        if height <= 0 or width <= 0:
            self._renderer.unset_color_mode(self._color)
            return

        # Step 1: clamp each user-supplied point to [x_min..x_max], [y_min..y_max]
        # and sort them by X
        clamped_pts = []
        for (px, py) in self.points:
            cx = min(max(px, self._x_min), self._x_max)
            cy = min(max(py, self._y_min), self._y_max)
            clamped_pts.append((cx, cy))
        sorted_pts = sorted(clamped_pts, key=lambda p: p[0])
        if not sorted_pts:
            self._renderer.unset_color_mode(self._color)
            return

        # For convenience
        x_min, x_max = self._x_min, self._x_max
        y_min, y_max = self._y_min, self._y_max
        if x_max == x_min:
            x_max = x_min + 1
        if y_max == y_min:
            y_max = y_min + 1

        # Step 2: Prepare an ASCII buffer
        spark_matrix = [[" " for _ in range(width)] for _ in range(height)]

        # Helper to get Y-value for a given real_x:
        def get_y_for_x(xf):
            """
            Returns a y-value for the given x-value (xf) according to the
            current mode ("interpolate" or "step"), while also forcing
            values outside the first/last point to use the first/last y.
            """

            # If no points are defined, default to 0
            if len(sorted_pts) == 0:
                return 0

            # If only a single point is defined, always return that y
            if len(sorted_pts) == 1:
                return sorted_pts[0][1]

            # Grab first and last points for clamping
            first_x, first_y = sorted_pts[0]
            last_x, last_y = sorted_pts[-1]

            # 1) If xf is less than the first point's x
            if xf <= first_x and self.mode != "off":
                return first_y

            # 2) If xf is greater than the last point's x
            if xf >= last_x and self.mode != "off":
                return last_y

            # 3) Otherwise, we are between the first and last point
            if self.mode == "step":
                # "STEP" => find which segment xf is in, use the left point’s y
                for i in range(len(sorted_pts) - 1):
                    x1, y1 = sorted_pts[i]
                    x2, y2 = sorted_pts[i + 1]
                    if xf < x2:
                        return y1
                return last_y  # fallback
            elif self.mode == "interpolate":
                # "INTERPOLATE" => piecewise linear between the two bounding points
                for i in range(len(sorted_pts) - 1):
                    x1, y1 = sorted_pts[i]
                    x2, y2 = sorted_pts[i + 1]
                    if x1 <= xf <= x2:
                        # fraction
                        dx = x2 - x1
                        if dx == 0:
                            return y1
                        ratio = (xf - x1) / dx
                        return y1 + ratio * (y2 - y1)
                return last_y  # fallback, though normally won't reach here
            else:
                # "OFF" => Just display the interpolate curve with default points
                x1, y1 = self._x_min, self._y_min
                x2, y2 = self._x_max, self._y_max
                if x1 <= xf <= x2:
                    # fraction
                    dx = x2 - x1
                    if dx == 0:
                        return y1
                    ratio = (xf - x1) / dx
                    return y1 + ratio * (y2 - y1)

        # Populate spark_matrix with '*'
        for x_index in range(width):
            # Map x_index => real_x in [x_min..x_max]
            real_x = x_min + (x_index / max(1, width-1))*(x_max - x_min)
            real_y = get_y_for_x(real_x)

            # Normalize real_y => row
            y_ratio = (real_y - y_min)/(y_max - y_min)
            row = height - 1 - int(round(y_ratio*(height-1)))
            if 0 <= row < height:
                spark_matrix[row][x_index] = "*"

        # Step 3: Render spark_matrix line by line
        for row_i in range(height):
            row_str = "".join(spark_matrix[row_i])
            # Draw at (base_y + row_i, base_x)
            self._renderer.draw_text(self, row_str, base_y + row_i, start_pos=0)


        self._renderer.unset_color_mode(self._color)
