#!/usr/bin/env python3
import subprocess
from functools import partial
import argparse
import sys
import os
import json

from voxl_utils_python.voxl_utils import is_voxl

DEBUG=False
if DEBUG:
    sys.path.append(os.path.normpath(os.path.join(os.path.dirname(__file__), "../python_packages")))

from voxl_elrs_tools.crsf import *
import voxl_elrs_tools.elrs_device
from voxl_elrs_tools.elrs_device import ELRSDeviceManager
from voxl_elrs_tools.manage_firmware import ELRSFirmwareManager, PlatformTypes
from voxl_elrs_tools.configurator import ELRSConfigData, ELRSConfigurator, config_prompt, config_wizard
import elrs_python.BFinitPassthrough as BFinitPassthrough
from voxl_elrs_wizard.py_cui_wizard import run_wizard

##################################################################
### Global tools/definitions
##################################################################

dev_info = None

def get_int(prompt):
    try:
        out = int(input(prompt))
    except:
        print("Failed to parse value")
        return get_int(prompt)
    return out

##################################################################
### Wizard Commands
##################################################################

def print_info():
    global dev_info
    dev_info = CRSFDev().run()
    if dev_info is not None and isinstance(dev_info, dict):
        print(f"Detected ELRS receiver named: {dev_info['name']} running ELRS version {dev_info['sw_ver']}")
        # print(f"Using ModalAI ELRS Firmware (Y/N)?: {'Y' if dev_info['modal_elrs'] else 'N'}")
    else:
        print("Failed to detect device info, are you sure it is connected properly?")
        dev_info = None


def bind():
    """
    Called from wizard or bind (-b) command.
    Puts a receiver into bind mode until it becomes connected to a transmitter.

    Flags
    ==============================
     -d: Serial Device
    """
    CRSFBind().run()
    print("Binding!")

def unbind():
    """
    Called from wizard or unbind (-u) command.
    Disconnects a bound receiver and sets and UID back to the configured value.
    
    Note: If connected with a binding phrase the tx/rx will rebind after a short time

    Flags
    ==============================
     -d: Serial Device
    """
    CRSFUnbind().run()
    print("Unbinding!")


def pwm_map():
    """
    Called from wizard
    Maps an RC channel to a PWM pin

    Arguments
    ============================
     in_ch: Channel number (this is based on the tx channels NOT the output pins)
     pin: PWM pin to use (1-4)

    Flags
    ==============================
     -d: Serial Device
    """
    print("Set the PWM parameters:")
    pin = get_int("Pin to set (0-3) > ")
    in_ch = get_int("Input Channel (1-255) > ")
    cmd = CRSFSetPWMCh(pin, in_ch)
    cmd.run()
    print("PWM mapping set")

def set_id(phrase=None):
    """
    Called from wizard
    Sets the binding phrase of a receiver

    Arguments
    ============================
     binding_phrase: the binding phrase to use

    Flags
    ==============================
     -d: Serial Device
    """
    if phrase is None or len(phrase) < 1:
        print("Enter the binding phrase or press enter to cancel")
        binding_phrase = input("> ")
        if len(binding_phrase) < 1:
            print("Cancelled!")
            return
    else:
        binding_phrase = phrase
    ret = CRSFSetID(binding_phrase).run()
    print("Set the binding phrase!") if ret == 0 else print("Failed to set the binding phrase")


def restart_receiver():
    """
    Called from wizard
    Restarts a receiver

    Flags
    ==============================
     -d: Serial Device
    """
    CRSFSetBootloader().run()
    print("Receiver Reset")


def flash_receiver(auto: bool = True, force=False):
    """
    Called from wizard
    Flashes a receiver

    :param `bool` auto: Automatically detect and update the firmware of the attached device 

    Flags
    ==============================
     -d: Serial Device
    """
    global dev_info
    if auto:
        dev_manager = ELRSDeviceManager()
        if dev_info is not None:
            firm = ELRSFirmwareManager(dev_manager)
            ret = firm.auto_update(name=dev_info["name"], min_ver=dev_info["sw_ver"])
            dev_info = CRSFDev(dev_manager).run()
        else:
            auto = False
            print("Cannot auto detect installed version, is it connected properly?")
    else:
        raise NotImplementedError("Use the wizard if auto-detection is necessary")


def flash_image(fw: dict = None):
    """
    Flashes a receiver with the provided firmware object

    :param `dict` fw: A description of the location and parameters of the firmware to upload

    Flags
    ==============================
     -d: Serial Device
    """
    dev_manager = ELRSDeviceManager()
    firm = ELRSFirmwareManager(dev_manager)
    if fw is not None:
        ret = firm.upload_firmware_from_dict(fw)
        dev_info = CRSFDev(dev_manager).run()
        if dev_info is not None and isinstance(dev_info, dict):
            print(f"Detected ELRS receiver named: {dev_info['name']} running ELRS version {dev_info['sw_ver']}")
            print(f"Using ModalAI ELRS Firmware (Y/N)?: {'Y' if dev_info['modal_elrs'] else 'N'}")


def flash_transmitter(transmitter: str = ""):
    """
    TODO: Work with generated binaries from modalai/ExpressLRS and support all transmitters
    Called from wizard
    Flashes a transmitter

    :param `str` transmitter: The name of the device to flash

    Flags
    ==============================
     -d: Serial Device
    """
    import esptool, time

    elrs_dev = ELRSDeviceManager()
    time.sleep(5)

    if transmitter == "IFlightCommando":
        tx_flash_command = ["--chip", "esp32", "--port", str(elrs_dev.device), "--baud", "460800", "--before", "default_reset", "--after", "hard_reset", "write_flash", "-z",
                            "--flash_mode", "dio", "--flash_freq", "40m", "--flash_size=keep",
                            "0x1000",  elrs_dev.TX_3_0_1[0],
                            "0x8000",  elrs_dev.TX_3_0_1[1],
                            "0xe000",  elrs_dev.TX_3_0_1[2],
                            "0x10000", elrs_dev.TX_3_0_1[3],]
    elif transmitter == "IFlightCommando2":
        tx_flash_command = ["--chip", "esp32", "--port", str(elrs_dev.device), "--baud", "460800", "--before", "default_reset", "--after", "hard_reset", "write_flash", "-z",
                            "--flash_mode", "dio", "--flash_freq", "40m", "--flash_size=keep",
                            "0x1000",  elrs_dev.TX_V2_3_4_2[0],
                            "0x8000",  elrs_dev.TX_V2_3_4_2[1],
                            "0xe000",  elrs_dev.TX_V2_3_4_2[2],
                            "0x10000", elrs_dev.TX_V2_3_4_2[3],]
        
    result = esptool.main(tx_flash_command)

    #print("Would you like to reformat a connected SD card and copy the ELRS Lua script? Y/N")
    #inval = input("> ")
    #if inval.lower() == "n" or inval.lower() == "no":
    #    return
    
    # TODO: Auto flash an SD CARD


def is_fw_old(current_version, expected_version):
    try:
        v1 = tuple(map(int, current_version.split(".")))
        v2 = tuple(map(int, expected_version.split(".")))
        return v1 < v2
    except ValueError:
        print("Invalid version format")
        return False


##################################################################
### Argument parser and main execution
##################################################################
    
parser = argparse.ArgumentParser(add_help=False)
# Parse args
def parse_args() -> dict:
    """
    Parse and validate args
    """
    parser.add_argument('command', help="Command to run (optional) [help, bind, unbind, update, info, setid, configure]", nargs='*', action="store")
    parser.add_argument('-h', '--help', help='Show this help message and exit', required=False, action='store_true')
    parser.add_argument('-d', '--device', help='Serial port', default='', required=False, type=str)
    parser.add_argument('--dir', help='Root directory for firmware', default='', required=False, type=str)
    parser.add_argument('-w', '--wizard', help='Activate interactive wizard', required=False, action='store_true')
    parser.add_argument('-q', '--quiet', help="Quiet mode, don't print debug msgs", required=False, action='store_true')
    parser.add_argument('-b', '--bind', help='Put receiver into bind mode', required=False, action='store_true')
    parser.add_argument('-B', '--betaflight', help='Put betaflight flight controller into passthrough mode', required=False, action='store_true')
    parser.add_argument('-u', '--unbind', help='Unbind a bound transmitter', required=False, action='store_true')
    parser.add_argument('-l', '--bootloader', help='Put receiver into bootloader mode', required=False, action='store_true')
    parser.add_argument('-v', '--version', help='Get receiver version info', required=False, action='store_true')
    parser.add_argument('-s', '--scan', help='Scan current firmware and update if <=3.2.1', required=False, action='store_true')
    parser.add_argument('-p', '--setid', help='Sets the binding phrase of a connected receiver', required=False, action='store')
    parser.add_argument('-n', '--noscan', help='Disables auto-detection of the connected receiver', required=False, action='store_true')
    return vars(parser.parse_args())

def main():
    args = parse_args()

    if is_voxl():
        import voxl_utils_python.modal_journal as journal
        # journal.modal_journal_set_level(journal.ModalJournalLevel.PRINT)

    # Check if any command is specified
    if args["command"] is None or len(args["command"]) < 1:
        command = ""
    else:
        command = args["command"][0]

    if args["help"] or command == "help":
        parser.print_help()
        exit()

    if len(args["dir"]) > 0:
        voxl_elrs_tools.elrs_device.elrs_dir = args["dir"]

    # If not help, then try and connect to the device
    if args["device"]:
        voxl_elrs_tools.elrs_device.elrs_default_device = args["device"]

    # Initialize betaflight passthrough here
    if args["betaflight"]:
        try:
            BFinitPassthrough.bf_passthrough_init(voxl_elrs_tools.elrs_device.elrs_default_device, 420000, False)
        except BFinitPassthrough.PassthroughEnabled as info:
            print("  Warning: '%s'\n" % info)
        except BFinitPassthrough.PassthroughFailed as failed:
            SystemExit(failed)

    if command == "txbind":
        send_tx_bind()
        exit()

    # Execute any commands passed
    if args["bind"] or command == "bind":
        bind()
        exit()

    if args["unbind"] or command == "unbind":
        unbind()
        exit()

    if args["scan"] or command == "update":
        print("Autodetecting device information...")
        dev_info = CRSFDev().run()

        # only flash receiver if fw is out of date
        if dev_info is not None and isinstance(dev_info, dict):
            flash_receiver(True)

        exit()

    if args["setid"] or command == "setid":
        # Check if a binding phrase was passed positionally
        parser.add_argument("pos_id", nargs='?', action="store")
        args = vars(parser.parse_args())
        if args["pos_id"] is not None and len(args["pos_id"]) > 0:
            set_id(args["pos_id"])
            print(args["pos_id"])
        elif args["setid"] is not None and len(args["setid"]) > 0:
            set_id(args["setid"])
            print(args["setid"])
        else:
            print("No binding phrase specified! Please use the format \"voxl-elrs setid <BINDING_PHRASE>\" Exiting...")
        exit()

    if args["version"] or command == "info":
        print_info()
        exit()

    if command == "configure":
        # Apply the following configuration
        #parser = argparse.ArgumentParser(add_help=False)
        #parser.add_argument('command', help="Command to run (optional) [help, bind, unbind, update, info, setid, configure]", action="store")
        #parser.add_argument("config", nargs='?', action="store")
        #conf_args = vars(parser.parse_args())
        if len(args["command"]) > 1:
            print(f"looking for configuration named '{args['command'][1]}.conf'...")
            if(os.path.isfile(f"/usr/share/modalai/voxl-elrs/config/{args['command'][1]}.conf")):
                conf = ELRSConfigurator()
                conf.load_data(file_name=f"/usr/share/modalai/voxl-elrs/config/{args['command'][1]}.conf")
                conf.save_data()
                conf.configure_default()
            else:
                print("Cannot find configuration file! Exiting...")

        else:
            print("Uploading config!")
            # Use default config
            ELRSConfigurator().configure_default()
        exit()

    if args["noscan"]:
        run_wizard(no_scan=True)
    else:
        run_wizard(no_scan=False)
